// Copyright 2021 The MediaPipe Authors.
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package com.google.mediapipe.solutions.facemesh;

import com.google.auto.value.AutoValue;
import com.google.common.collect.ImmutableSet;

/** FaceMesh connections constants. */
public final class FaceMeshConnections {

  /** Value class representing hand connection. */
  @AutoValue
  public abstract static class Connection {
    static Connection create(int start, int end) {
      return new AutoValue_FaceMeshConnections_Connection(start, end);
    }

    public abstract int start();

    public abstract int end();
  }

  public static final ImmutableSet<Connection> FACEMESH_LIPS =
      ImmutableSet.of(
          Connection.create(61, 146),
          Connection.create(146, 91),
          Connection.create(91, 181),
          Connection.create(181, 84),
          Connection.create(84, 17),
          Connection.create(17, 314),
          Connection.create(314, 405),
          Connection.create(405, 321),
          Connection.create(321, 375),
          Connection.create(375, 291),
          Connection.create(61, 185),
          Connection.create(185, 40),
          Connection.create(40, 39),
          Connection.create(39, 37),
          Connection.create(37, 0),
          Connection.create(0, 267),
          Connection.create(267, 269),
          Connection.create(269, 270),
          Connection.create(270, 409),
          Connection.create(409, 291),
          Connection.create(78, 95),
          Connection.create(95, 88),
          Connection.create(88, 178),
          Connection.create(178, 87),
          Connection.create(87, 14),
          Connection.create(14, 317),
          Connection.create(317, 402),
          Connection.create(402, 318),
          Connection.create(318, 324),
          Connection.create(324, 308),
          Connection.create(78, 191),
          Connection.create(191, 80),
          Connection.create(80, 81),
          Connection.create(81, 82),
          Connection.create(82, 13),
          Connection.create(13, 312),
          Connection.create(312, 311),
          Connection.create(311, 310),
          Connection.create(310, 415),
          Connection.create(415, 308));

  public static final ImmutableSet<Connection> FACEMESH_LEFT_EYE =
      ImmutableSet.of(
          Connection.create(263, 249),
          Connection.create(249, 390),
          Connection.create(390, 373),
          Connection.create(373, 374),
          Connection.create(374, 380),
          Connection.create(380, 381),
          Connection.create(381, 382),
          Connection.create(382, 362),
          Connection.create(263, 466),
          Connection.create(466, 388),
          Connection.create(388, 387),
          Connection.create(387, 386),
          Connection.create(386, 385),
          Connection.create(385, 384),
          Connection.create(384, 398),
          Connection.create(398, 362));

  public static final ImmutableSet<Connection> FACEMESH_LEFT_EYEBROW =
      ImmutableSet.of(
          Connection.create(276, 283),
          Connection.create(283, 282),
          Connection.create(282, 295),
          Connection.create(295, 285),
          Connection.create(300, 293),
          Connection.create(293, 334),
          Connection.create(334, 296),
          Connection.create(296, 336));

  public static final ImmutableSet<Connection> FACEMESH_LEFT_IRIS =
      ImmutableSet.of(
          Connection.create(474, 475),
          Connection.create(475, 476),
          Connection.create(476, 477),
          Connection.create(477, 474));

  public static final ImmutableSet<Connection> FACEMESH_RIGHT_EYE =
      ImmutableSet.of(
          Connection.create(33, 7),
          Connection.create(7, 163),
          Connection.create(163, 144),
          Connection.create(144, 145),
          Connection.create(145, 153),
          Connection.create(153, 154),
          Connection.create(154, 155),
          Connection.create(155, 133),
          Connection.create(33, 246),
          Connection.create(246, 161),
          Connection.create(161, 160),
          Connection.create(160, 159),
          Connection.create(159, 158),
          Connection.create(158, 157),
          Connection.create(157, 173),
          Connection.create(173, 133));

  public static final ImmutableSet<Connection> FACEMESH_RIGHT_EYEBROW =
      ImmutableSet.of(
          Connection.create(46, 53),
          Connection.create(53, 52),
          Connection.create(52, 65),
          Connection.create(65, 55),
          Connection.create(70, 63),
          Connection.create(63, 105),
          Connection.create(105, 66),
          Connection.create(66, 107));

  public static final ImmutableSet<Connection> FACEMESH_RIGHT_IRIS =
      ImmutableSet.of(
          Connection.create(469, 470),
          Connection.create(470, 471),
          Connection.create(471, 472),
          Connection.create(472, 469));

  public static final ImmutableSet<Connection> FACEMESH_FACE_OVAL =
      ImmutableSet.of(
          Connection.create(10, 338),
          Connection.create(338, 297),
          Connection.create(297, 332),
          Connection.create(332, 284),
          Connection.create(284, 251),
          Connection.create(251, 389),
          Connection.create(389, 356),
          Connection.create(356, 454),
          Connection.create(454, 323),
          Connection.create(323, 361),
          Connection.create(361, 288),
          Connection.create(288, 397),
          Connection.create(397, 365),
          Connection.create(365, 379),
          Connection.create(379, 378),
          Connection.create(378, 400),
          Connection.create(400, 377),
          Connection.create(377, 152),
          Connection.create(152, 148),
          Connection.create(148, 176),
          Connection.create(176, 149),
          Connection.create(149, 150),
          Connection.create(150, 136),
          Connection.create(136, 172),
          Connection.create(172, 58),
          Connection.create(58, 132),
          Connection.create(132, 93),
          Connection.create(93, 234),
          Connection.create(234, 127),
          Connection.create(127, 162),
          Connection.create(162, 21),
          Connection.create(21, 54),
          Connection.create(54, 103),
          Connection.create(103, 67),
          Connection.create(67, 109),
          Connection.create(109, 10));

  public static final ImmutableSet<Connection> FACEMESH_CONTOURS =
      ImmutableSet.<Connection>builder()
          .addAll(FACEMESH_LIPS)
          .addAll(FACEMESH_LEFT_EYE)
          .addAll(FACEMESH_LEFT_EYEBROW)
          .addAll(FACEMESH_RIGHT_EYE)
          .addAll(FACEMESH_RIGHT_EYEBROW)
          .addAll(FACEMESH_FACE_OVAL)
          .build();

  public static final ImmutableSet<Connection> FACEMESH_TESSELATION =
      ImmutableSet.of(
          Connection.create(127, 34),
          Connection.create(34, 139),
          Connection.create(139, 127),
          Connection.create(11, 0),
          Connection.create(0, 37),
          Connection.create(37, 11),
          Connection.create(232, 231),
          Connection.create(231, 120),
          Connection.create(120, 232),
          Connection.create(72, 37),
          Connection.create(37, 39),
          Connection.create(39, 72),
          Connection.create(128, 121),
          Connection.create(121, 47),
          Connection.create(47, 128),
          Connection.create(232, 121),
          Connection.create(121, 128),
          Connection.create(128, 232),
          Connection.create(104, 69),
          Connection.create(69, 67),
          Connection.create(67, 104),
          Connection.create(175, 171),
          Connection.create(171, 148),
          Connection.create(148, 175),
          Connection.create(118, 50),
          Connection.create(50, 101),
          Connection.create(101, 118),
          Connection.create(73, 39),
          Connection.create(39, 40),
          Connection.create(40, 73),
          Connection.create(9, 151),
          Connection.create(151, 108),
          Connection.create(108, 9),
          Connection.create(48, 115),
          Connection.create(115, 131),
          Connection.create(131, 48),
          Connection.create(194, 204),
          Connection.create(204, 211),
          Connection.create(211, 194),
          Connection.create(74, 40),
          Connection.create(40, 185),
          Connection.create(185, 74),
          Connection.create(80, 42),
          Connection.create(42, 183),
          Connection.create(183, 80),
          Connection.create(40, 92),
          Connection.create(92, 186),
          Connection.create(186, 40),
          Connection.create(230, 229),
          Connection.create(229, 118),
          Connection.create(118, 230),
          Connection.create(202, 212),
          Connection.create(212, 214),
          Connection.create(214, 202),
          Connection.create(83, 18),
          Connection.create(18, 17),
          Connection.create(17, 83),
          Connection.create(76, 61),
          Connection.create(61, 146),
          Connection.create(146, 76),
          Connection.create(160, 29),
          Connection.create(29, 30),
          Connection.create(30, 160),
          Connection.create(56, 157),
          Connection.create(157, 173),
          Connection.create(173, 56),
          Connection.create(106, 204),
          Connection.create(204, 194),
          Connection.create(194, 106),
          Connection.create(135, 214),
          Connection.create(214, 192),
          Connection.create(192, 135),
          Connection.create(203, 165),
          Connection.create(165, 98),
          Connection.create(98, 203),
          Connection.create(21, 71),
          Connection.create(71, 68),
          Connection.create(68, 21),
          Connection.create(51, 45),
          Connection.create(45, 4),
          Connection.create(4, 51),
          Connection.create(144, 24),
          Connection.create(24, 23),
          Connection.create(23, 144),
          Connection.create(77, 146),
          Connection.create(146, 91),
          Connection.create(91, 77),
          Connection.create(205, 50),
          Connection.create(50, 187),
          Connection.create(187, 205),
          Connection.create(201, 200),
          Connection.create(200, 18),
          Connection.create(18, 201),
          Connection.create(91, 106),
          Connection.create(106, 182),
          Connection.create(182, 91),
          Connection.create(90, 91),
          Connection.create(91, 181),
          Connection.create(181, 90),
          Connection.create(85, 84),
          Connection.create(84, 17),
          Connection.create(17, 85),
          Connection.create(206, 203),
          Connection.create(203, 36),
          Connection.create(36, 206),
          Connection.create(148, 171),
          Connection.create(171, 140),
          Connection.create(140, 148),
          Connection.create(92, 40),
          Connection.create(40, 39),
          Connection.create(39, 92),
          Connection.create(193, 189),
          Connection.create(189, 244),
          Connection.create(244, 193),
          Connection.create(159, 158),
          Connection.create(158, 28),
          Connection.create(28, 159),
          Connection.create(247, 246),
          Connection.create(246, 161),
          Connection.create(161, 247),
          Connection.create(236, 3),
          Connection.create(3, 196),
          Connection.create(196, 236),
          Connection.create(54, 68),
          Connection.create(68, 104),
          Connection.create(104, 54),
          Connection.create(193, 168),
          Connection.create(168, 8),
          Connection.create(8, 193),
          Connection.create(117, 228),
          Connection.create(228, 31),
          Connection.create(31, 117),
          Connection.create(189, 193),
          Connection.create(193, 55),
          Connection.create(55, 189),
          Connection.create(98, 97),
          Connection.create(97, 99),
          Connection.create(99, 98),
          Connection.create(126, 47),
          Connection.create(47, 100),
          Connection.create(100, 126),
          Connection.create(166, 79),
          Connection.create(79, 218),
          Connection.create(218, 166),
          Connection.create(155, 154),
          Connection.create(154, 26),
          Connection.create(26, 155),
          Connection.create(209, 49),
          Connection.create(49, 131),
          Connection.create(131, 209),
          Connection.create(135, 136),
          Connection.create(136, 150),
          Connection.create(150, 135),
          Connection.create(47, 126),
          Connection.create(126, 217),
          Connection.create(217, 47),
          Connection.create(223, 52),
          Connection.create(52, 53),
          Connection.create(53, 223),
          Connection.create(45, 51),
          Connection.create(51, 134),
          Connection.create(134, 45),
          Connection.create(211, 170),
          Connection.create(170, 140),
          Connection.create(140, 211),
          Connection.create(67, 69),
          Connection.create(69, 108),
          Connection.create(108, 67),
          Connection.create(43, 106),
          Connection.create(106, 91),
          Connection.create(91, 43),
          Connection.create(230, 119),
          Connection.create(119, 120),
          Connection.create(120, 230),
          Connection.create(226, 130),
          Connection.create(130, 247),
          Connection.create(247, 226),
          Connection.create(63, 53),
          Connection.create(53, 52),
          Connection.create(52, 63),
          Connection.create(238, 20),
          Connection.create(20, 242),
          Connection.create(242, 238),
          Connection.create(46, 70),
          Connection.create(70, 156),
          Connection.create(156, 46),
          Connection.create(78, 62),
          Connection.create(62, 96),
          Connection.create(96, 78),
          Connection.create(46, 53),
          Connection.create(53, 63),
          Connection.create(63, 46),
          Connection.create(143, 34),
          Connection.create(34, 227),
          Connection.create(227, 143),
          Connection.create(123, 117),
          Connection.create(117, 111),
          Connection.create(111, 123),
          Connection.create(44, 125),
          Connection.create(125, 19),
          Connection.create(19, 44),
          Connection.create(236, 134),
          Connection.create(134, 51),
          Connection.create(51, 236),
          Connection.create(216, 206),
          Connection.create(206, 205),
          Connection.create(205, 216),
          Connection.create(154, 153),
          Connection.create(153, 22),
          Connection.create(22, 154),
          Connection.create(39, 37),
          Connection.create(37, 167),
          Connection.create(167, 39),
          Connection.create(200, 201),
          Connection.create(201, 208),
          Connection.create(208, 200),
          Connection.create(36, 142),
          Connection.create(142, 100),
          Connection.create(100, 36),
          Connection.create(57, 212),
          Connection.create(212, 202),
          Connection.create(202, 57),
          Connection.create(20, 60),
          Connection.create(60, 99),
          Connection.create(99, 20),
          Connection.create(28, 158),
          Connection.create(158, 157),
          Connection.create(157, 28),
          Connection.create(35, 226),
          Connection.create(226, 113),
          Connection.create(113, 35),
          Connection.create(160, 159),
          Connection.create(159, 27),
          Connection.create(27, 160),
          Connection.create(204, 202),
          Connection.create(202, 210),
          Connection.create(210, 204),
          Connection.create(113, 225),
          Connection.create(225, 46),
          Connection.create(46, 113),
          Connection.create(43, 202),
          Connection.create(202, 204),
          Connection.create(204, 43),
          Connection.create(62, 76),
          Connection.create(76, 77),
          Connection.create(77, 62),
          Connection.create(137, 123),
          Connection.create(123, 116),
          Connection.create(116, 137),
          Connection.create(41, 38),
          Connection.create(38, 72),
          Connection.create(72, 41),
          Connection.create(203, 129),
          Connection.create(129, 142),
          Connection.create(142, 203),
          Connection.create(64, 98),
          Connection.create(98, 240),
          Connection.create(240, 64),
          Connection.create(49, 102),
          Connection.create(102, 64),
          Connection.create(64, 49),
          Connection.create(41, 73),
          Connection.create(73, 74),
          Connection.create(74, 41),
          Connection.create(212, 216),
          Connection.create(216, 207),
          Connection.create(207, 212),
          Connection.create(42, 74),
          Connection.create(74, 184),
          Connection.create(184, 42),
          Connection.create(169, 170),
          Connection.create(170, 211),
          Connection.create(211, 169),
          Connection.create(170, 149),
          Connection.create(149, 176),
          Connection.create(176, 170),
          Connection.create(105, 66),
          Connection.create(66, 69),
          Connection.create(69, 105),
          Connection.create(122, 6),
          Connection.create(6, 168),
          Connection.create(168, 122),
          Connection.create(123, 147),
          Connection.create(147, 187),
          Connection.create(187, 123),
          Connection.create(96, 77),
          Connection.create(77, 90),
          Connection.create(90, 96),
          Connection.create(65, 55),
          Connection.create(55, 107),
          Connection.create(107, 65),
          Connection.create(89, 90),
          Connection.create(90, 180),
          Connection.create(180, 89),
          Connection.create(101, 100),
          Connection.create(100, 120),
          Connection.create(120, 101),
          Connection.create(63, 105),
          Connection.create(105, 104),
          Connection.create(104, 63),
          Connection.create(93, 137),
          Connection.create(137, 227),
          Connection.create(227, 93),
          Connection.create(15, 86),
          Connection.create(86, 85),
          Connection.create(85, 15),
          Connection.create(129, 102),
          Connection.create(102, 49),
          Connection.create(49, 129),
          Connection.create(14, 87),
          Connection.create(87, 86),
          Connection.create(86, 14),
          Connection.create(55, 8),
          Connection.create(8, 9),
          Connection.create(9, 55),
          Connection.create(100, 47),
          Connection.create(47, 121),
          Connection.create(121, 100),
          Connection.create(145, 23),
          Connection.create(23, 22),
          Connection.create(22, 145),
          Connection.create(88, 89),
          Connection.create(89, 179),
          Connection.create(179, 88),
          Connection.create(6, 122),
          Connection.create(122, 196),
          Connection.create(196, 6),
          Connection.create(88, 95),
          Connection.create(95, 96),
          Connection.create(96, 88),
          Connection.create(138, 172),
          Connection.create(172, 136),
          Connection.create(136, 138),
          Connection.create(215, 58),
          Connection.create(58, 172),
          Connection.create(172, 215),
          Connection.create(115, 48),
          Connection.create(48, 219),
          Connection.create(219, 115),
          Connection.create(42, 80),
          Connection.create(80, 81),
          Connection.create(81, 42),
          Connection.create(195, 3),
          Connection.create(3, 51),
          Connection.create(51, 195),
          Connection.create(43, 146),
          Connection.create(146, 61),
          Connection.create(61, 43),
          Connection.create(171, 175),
          Connection.create(175, 199),
          Connection.create(199, 171),
          Connection.create(81, 82),
          Connection.create(82, 38),
          Connection.create(38, 81),
          Connection.create(53, 46),
          Connection.create(46, 225),
          Connection.create(225, 53),
          Connection.create(144, 163),
          Connection.create(163, 110),
          Connection.create(110, 144),
          Connection.create(52, 65),
          Connection.create(65, 66),
          Connection.create(66, 52),
          Connection.create(229, 228),
          Connection.create(228, 117),
          Connection.create(117, 229),
          Connection.create(34, 127),
          Connection.create(127, 234),
          Connection.create(234, 34),
          Connection.create(107, 108),
          Connection.create(108, 69),
          Connection.create(69, 107),
          Connection.create(109, 108),
          Connection.create(108, 151),
          Connection.create(151, 109),
          Connection.create(48, 64),
          Connection.create(64, 235),
          Connection.create(235, 48),
          Connection.create(62, 78),
          Connection.create(78, 191),
          Connection.create(191, 62),
          Connection.create(129, 209),
          Connection.create(209, 126),
          Connection.create(126, 129),
          Connection.create(111, 35),
          Connection.create(35, 143),
          Connection.create(143, 111),
          Connection.create(117, 123),
          Connection.create(123, 50),
          Connection.create(50, 117),
          Connection.create(222, 65),
          Connection.create(65, 52),
          Connection.create(52, 222),
          Connection.create(19, 125),
          Connection.create(125, 141),
          Connection.create(141, 19),
          Connection.create(221, 55),
          Connection.create(55, 65),
          Connection.create(65, 221),
          Connection.create(3, 195),
          Connection.create(195, 197),
          Connection.create(197, 3),
          Connection.create(25, 7),
          Connection.create(7, 33),
          Connection.create(33, 25),
          Connection.create(220, 237),
          Connection.create(237, 44),
          Connection.create(44, 220),
          Connection.create(70, 71),
          Connection.create(71, 139),
          Connection.create(139, 70),
          Connection.create(122, 193),
          Connection.create(193, 245),
          Connection.create(245, 122),
          Connection.create(247, 130),
          Connection.create(130, 33),
          Connection.create(33, 247),
          Connection.create(71, 21),
          Connection.create(21, 162),
          Connection.create(162, 71),
          Connection.create(170, 169),
          Connection.create(169, 150),
          Connection.create(150, 170),
          Connection.create(188, 174),
          Connection.create(174, 196),
          Connection.create(196, 188),
          Connection.create(216, 186),
          Connection.create(186, 92),
          Connection.create(92, 216),
          Connection.create(2, 97),
          Connection.create(97, 167),
          Connection.create(167, 2),
          Connection.create(141, 125),
          Connection.create(125, 241),
          Connection.create(241, 141),
          Connection.create(164, 167),
          Connection.create(167, 37),
          Connection.create(37, 164),
          Connection.create(72, 38),
          Connection.create(38, 12),
          Connection.create(12, 72),
          Connection.create(38, 82),
          Connection.create(82, 13),
          Connection.create(13, 38),
          Connection.create(63, 68),
          Connection.create(68, 71),
          Connection.create(71, 63),
          Connection.create(226, 35),
          Connection.create(35, 111),
          Connection.create(111, 226),
          Connection.create(101, 50),
          Connection.create(50, 205),
          Connection.create(205, 101),
          Connection.create(206, 92),
          Connection.create(92, 165),
          Connection.create(165, 206),
          Connection.create(209, 198),
          Connection.create(198, 217),
          Connection.create(217, 209),
          Connection.create(165, 167),
          Connection.create(167, 97),
          Connection.create(97, 165),
          Connection.create(220, 115),
          Connection.create(115, 218),
          Connection.create(218, 220),
          Connection.create(133, 112),
          Connection.create(112, 243),
          Connection.create(243, 133),
          Connection.create(239, 238),
          Connection.create(238, 241),
          Connection.create(241, 239),
          Connection.create(214, 135),
          Connection.create(135, 169),
          Connection.create(169, 214),
          Connection.create(190, 173),
          Connection.create(173, 133),
          Connection.create(133, 190),
          Connection.create(171, 208),
          Connection.create(208, 32),
          Connection.create(32, 171),
          Connection.create(125, 44),
          Connection.create(44, 237),
          Connection.create(237, 125),
          Connection.create(86, 87),
          Connection.create(87, 178),
          Connection.create(178, 86),
          Connection.create(85, 86),
          Connection.create(86, 179),
          Connection.create(179, 85),
          Connection.create(84, 85),
          Connection.create(85, 180),
          Connection.create(180, 84),
          Connection.create(83, 84),
          Connection.create(84, 181),
          Connection.create(181, 83),
          Connection.create(201, 83),
          Connection.create(83, 182),
          Connection.create(182, 201),
          Connection.create(137, 93),
          Connection.create(93, 132),
          Connection.create(132, 137),
          Connection.create(76, 62),
          Connection.create(62, 183),
          Connection.create(183, 76),
          Connection.create(61, 76),
          Connection.create(76, 184),
          Connection.create(184, 61),
          Connection.create(57, 61),
          Connection.create(61, 185),
          Connection.create(185, 57),
          Connection.create(212, 57),
          Connection.create(57, 186),
          Connection.create(186, 212),
          Connection.create(214, 207),
          Connection.create(207, 187),
          Connection.create(187, 214),
          Connection.create(34, 143),
          Connection.create(143, 156),
          Connection.create(156, 34),
          Connection.create(79, 239),
          Connection.create(239, 237),
          Connection.create(237, 79),
          Connection.create(123, 137),
          Connection.create(137, 177),
          Connection.create(177, 123),
          Connection.create(44, 1),
          Connection.create(1, 4),
          Connection.create(4, 44),
          Connection.create(201, 194),
          Connection.create(194, 32),
          Connection.create(32, 201),
          Connection.create(64, 102),
          Connection.create(102, 129),
          Connection.create(129, 64),
          Connection.create(213, 215),
          Connection.create(215, 138),
          Connection.create(138, 213),
          Connection.create(59, 166),
          Connection.create(166, 219),
          Connection.create(219, 59),
          Connection.create(242, 99),
          Connection.create(99, 97),
          Connection.create(97, 242),
          Connection.create(2, 94),
          Connection.create(94, 141),
          Connection.create(141, 2),
          Connection.create(75, 59),
          Connection.create(59, 235),
          Connection.create(235, 75),
          Connection.create(24, 110),
          Connection.create(110, 228),
          Connection.create(228, 24),
          Connection.create(25, 130),
          Connection.create(130, 226),
          Connection.create(226, 25),
          Connection.create(23, 24),
          Connection.create(24, 229),
          Connection.create(229, 23),
          Connection.create(22, 23),
          Connection.create(23, 230),
          Connection.create(230, 22),
          Connection.create(26, 22),
          Connection.create(22, 231),
          Connection.create(231, 26),
          Connection.create(112, 26),
          Connection.create(26, 232),
          Connection.create(232, 112),
          Connection.create(189, 190),
          Connection.create(190, 243),
          Connection.create(243, 189),
          Connection.create(221, 56),
          Connection.create(56, 190),
          Connection.create(190, 221),
          Connection.create(28, 56),
          Connection.create(56, 221),
          Connection.create(221, 28),
          Connection.create(27, 28),
          Connection.create(28, 222),
          Connection.create(222, 27),
          Connection.create(29, 27),
          Connection.create(27, 223),
          Connection.create(223, 29),
          Connection.create(30, 29),
          Connection.create(29, 224),
          Connection.create(224, 30),
          Connection.create(247, 30),
          Connection.create(30, 225),
          Connection.create(225, 247),
          Connection.create(238, 79),
          Connection.create(79, 20),
          Connection.create(20, 238),
          Connection.create(166, 59),
          Connection.create(59, 75),
          Connection.create(75, 166),
          Connection.create(60, 75),
          Connection.create(75, 240),
          Connection.create(240, 60),
          Connection.create(147, 177),
          Connection.create(177, 215),
          Connection.create(215, 147),
          Connection.create(20, 79),
          Connection.create(79, 166),
          Connection.create(166, 20),
          Connection.create(187, 147),
          Connection.create(147, 213),
          Connection.create(213, 187),
          Connection.create(112, 233),
          Connection.create(233, 244),
          Connection.create(244, 112),
          Connection.create(233, 128),
          Connection.create(128, 245),
          Connection.create(245, 233),
          Connection.create(128, 114),
          Connection.create(114, 188),
          Connection.create(188, 128),
          Connection.create(114, 217),
          Connection.create(217, 174),
          Connection.create(174, 114),
          Connection.create(131, 115),
          Connection.create(115, 220),
          Connection.create(220, 131),
          Connection.create(217, 198),
          Connection.create(198, 236),
          Connection.create(236, 217),
          Connection.create(198, 131),
          Connection.create(131, 134),
          Connection.create(134, 198),
          Connection.create(177, 132),
          Connection.create(132, 58),
          Connection.create(58, 177),
          Connection.create(143, 35),
          Connection.create(35, 124),
          Connection.create(124, 143),
          Connection.create(110, 163),
          Connection.create(163, 7),
          Connection.create(7, 110),
          Connection.create(228, 110),
          Connection.create(110, 25),
          Connection.create(25, 228),
          Connection.create(356, 389),
          Connection.create(389, 368),
          Connection.create(368, 356),
          Connection.create(11, 302),
          Connection.create(302, 267),
          Connection.create(267, 11),
          Connection.create(452, 350),
          Connection.create(350, 349),
          Connection.create(349, 452),
          Connection.create(302, 303),
          Connection.create(303, 269),
          Connection.create(269, 302),
          Connection.create(357, 343),
          Connection.create(343, 277),
          Connection.create(277, 357),
          Connection.create(452, 453),
          Connection.create(453, 357),
          Connection.create(357, 452),
          Connection.create(333, 332),
          Connection.create(332, 297),
          Connection.create(297, 333),
          Connection.create(175, 152),
          Connection.create(152, 377),
          Connection.create(377, 175),
          Connection.create(347, 348),
          Connection.create(348, 330),
          Connection.create(330, 347),
          Connection.create(303, 304),
          Connection.create(304, 270),
          Connection.create(270, 303),
          Connection.create(9, 336),
          Connection.create(336, 337),
          Connection.create(337, 9),
          Connection.create(278, 279),
          Connection.create(279, 360),
          Connection.create(360, 278),
          Connection.create(418, 262),
          Connection.create(262, 431),
          Connection.create(431, 418),
          Connection.create(304, 408),
          Connection.create(408, 409),
          Connection.create(409, 304),
          Connection.create(310, 415),
          Connection.create(415, 407),
          Connection.create(407, 310),
          Connection.create(270, 409),
          Connection.create(409, 410),
          Connection.create(410, 270),
          Connection.create(450, 348),
          Connection.create(348, 347),
          Connection.create(347, 450),
          Connection.create(422, 430),
          Connection.create(430, 434),
          Connection.create(434, 422),
          Connection.create(313, 314),
          Connection.create(314, 17),
          Connection.create(17, 313),
          Connection.create(306, 307),
          Connection.create(307, 375),
          Connection.create(375, 306),
          Connection.create(387, 388),
          Connection.create(388, 260),
          Connection.create(260, 387),
          Connection.create(286, 414),
          Connection.create(414, 398),
          Connection.create(398, 286),
          Connection.create(335, 406),
          Connection.create(406, 418),
          Connection.create(418, 335),
          Connection.create(364, 367),
          Connection.create(367, 416),
          Connection.create(416, 364),
          Connection.create(423, 358),
          Connection.create(358, 327),
          Connection.create(327, 423),
          Connection.create(251, 284),
          Connection.create(284, 298),
          Connection.create(298, 251),
          Connection.create(281, 5),
          Connection.create(5, 4),
          Connection.create(4, 281),
          Connection.create(373, 374),
          Connection.create(374, 253),
          Connection.create(253, 373),
          Connection.create(307, 320),
          Connection.create(320, 321),
          Connection.create(321, 307),
          Connection.create(425, 427),
          Connection.create(427, 411),
          Connection.create(411, 425),
          Connection.create(421, 313),
          Connection.create(313, 18),
          Connection.create(18, 421),
          Connection.create(321, 405),
          Connection.create(405, 406),
          Connection.create(406, 321),
          Connection.create(320, 404),
          Connection.create(404, 405),
          Connection.create(405, 320),
          Connection.create(315, 16),
          Connection.create(16, 17),
          Connection.create(17, 315),
          Connection.create(426, 425),
          Connection.create(425, 266),
          Connection.create(266, 426),
          Connection.create(377, 400),
          Connection.create(400, 369),
          Connection.create(369, 377),
          Connection.create(322, 391),
          Connection.create(391, 269),
          Connection.create(269, 322),
          Connection.create(417, 465),
          Connection.create(465, 464),
          Connection.create(464, 417),
          Connection.create(386, 257),
          Connection.create(257, 258),
          Connection.create(258, 386),
          Connection.create(466, 260),
          Connection.create(260, 388),
          Connection.create(388, 466),
          Connection.create(456, 399),
          Connection.create(399, 419),
          Connection.create(419, 456),
          Connection.create(284, 332),
          Connection.create(332, 333),
          Connection.create(333, 284),
          Connection.create(417, 285),
          Connection.create(285, 8),
          Connection.create(8, 417),
          Connection.create(346, 340),
          Connection.create(340, 261),
          Connection.create(261, 346),
          Connection.create(413, 441),
          Connection.create(441, 285),
          Connection.create(285, 413),
          Connection.create(327, 460),
          Connection.create(460, 328),
          Connection.create(328, 327),
          Connection.create(355, 371),
          Connection.create(371, 329),
          Connection.create(329, 355),
          Connection.create(392, 439),
          Connection.create(439, 438),
          Connection.create(438, 392),
          Connection.create(382, 341),
          Connection.create(341, 256),
          Connection.create(256, 382),
          Connection.create(429, 420),
          Connection.create(420, 360),
          Connection.create(360, 429),
          Connection.create(364, 394),
          Connection.create(394, 379),
          Connection.create(379, 364),
          Connection.create(277, 343),
          Connection.create(343, 437),
          Connection.create(437, 277),
          Connection.create(443, 444),
          Connection.create(444, 283),
          Connection.create(283, 443),
          Connection.create(275, 440),
          Connection.create(440, 363),
          Connection.create(363, 275),
          Connection.create(431, 262),
          Connection.create(262, 369),
          Connection.create(369, 431),
          Connection.create(297, 338),
          Connection.create(338, 337),
          Connection.create(337, 297),
          Connection.create(273, 375),
          Connection.create(375, 321),
          Connection.create(321, 273),
          Connection.create(450, 451),
          Connection.create(451, 349),
          Connection.create(349, 450),
          Connection.create(446, 342),
          Connection.create(342, 467),
          Connection.create(467, 446),
          Connection.create(293, 334),
          Connection.create(334, 282),
          Connection.create(282, 293),
          Connection.create(458, 461),
          Connection.create(461, 462),
          Connection.create(462, 458),
          Connection.create(276, 353),
          Connection.create(353, 383),
          Connection.create(383, 276),
          Connection.create(308, 324),
          Connection.create(324, 325),
          Connection.create(325, 308),
          Connection.create(276, 300),
          Connection.create(300, 293),
          Connection.create(293, 276),
          Connection.create(372, 345),
          Connection.create(345, 447),
          Connection.create(447, 372),
          Connection.create(352, 345),
          Connection.create(345, 340),
          Connection.create(340, 352),
          Connection.create(274, 1),
          Connection.create(1, 19),
          Connection.create(19, 274),
          Connection.create(456, 248),
          Connection.create(248, 281),
          Connection.create(281, 456),
          Connection.create(436, 427),
          Connection.create(427, 425),
          Connection.create(425, 436),
          Connection.create(381, 256),
          Connection.create(256, 252),
          Connection.create(252, 381),
          Connection.create(269, 391),
          Connection.create(391, 393),
          Connection.create(393, 269),
          Connection.create(200, 199),
          Connection.create(199, 428),
          Connection.create(428, 200),
          Connection.create(266, 330),
          Connection.create(330, 329),
          Connection.create(329, 266),
          Connection.create(287, 273),
          Connection.create(273, 422),
          Connection.create(422, 287),
          Connection.create(250, 462),
          Connection.create(462, 328),
          Connection.create(328, 250),
          Connection.create(258, 286),
          Connection.create(286, 384),
          Connection.create(384, 258),
          Connection.create(265, 353),
          Connection.create(353, 342),
          Connection.create(342, 265),
          Connection.create(387, 259),
          Connection.create(259, 257),
          Connection.create(257, 387),
          Connection.create(424, 431),
          Connection.create(431, 430),
          Connection.create(430, 424),
          Connection.create(342, 353),
          Connection.create(353, 276),
          Connection.create(276, 342),
          Connection.create(273, 335),
          Connection.create(335, 424),
          Connection.create(424, 273),
          Connection.create(292, 325),
          Connection.create(325, 307),
          Connection.create(307, 292),
          Connection.create(366, 447),
          Connection.create(447, 345),
          Connection.create(345, 366),
          Connection.create(271, 303),
          Connection.create(303, 302),
          Connection.create(302, 271),
          Connection.create(423, 266),
          Connection.create(266, 371),
          Connection.create(371, 423),
          Connection.create(294, 455),
          Connection.create(455, 460),
          Connection.create(460, 294),
          Connection.create(279, 278),
          Connection.create(278, 294),
          Connection.create(294, 279),
          Connection.create(271, 272),
          Connection.create(272, 304),
          Connection.create(304, 271),
          Connection.create(432, 434),
          Connection.create(434, 427),
          Connection.create(427, 432),
          Connection.create(272, 407),
          Connection.create(407, 408),
          Connection.create(408, 272),
          Connection.create(394, 430),
          Connection.create(430, 431),
          Connection.create(431, 394),
          Connection.create(395, 369),
          Connection.create(369, 400),
          Connection.create(400, 395),
          Connection.create(334, 333),
          Connection.create(333, 299),
          Connection.create(299, 334),
          Connection.create(351, 417),
          Connection.create(417, 168),
          Connection.create(168, 351),
          Connection.create(352, 280),
          Connection.create(280, 411),
          Connection.create(411, 352),
          Connection.create(325, 319),
          Connection.create(319, 320),
          Connection.create(320, 325),
          Connection.create(295, 296),
          Connection.create(296, 336),
          Connection.create(336, 295),
          Connection.create(319, 403),
          Connection.create(403, 404),
          Connection.create(404, 319),
          Connection.create(330, 348),
          Connection.create(348, 349),
          Connection.create(349, 330),
          Connection.create(293, 298),
          Connection.create(298, 333),
          Connection.create(333, 293),
          Connection.create(323, 454),
          Connection.create(454, 447),
          Connection.create(447, 323),
          Connection.create(15, 16),
          Connection.create(16, 315),
          Connection.create(315, 15),
          Connection.create(358, 429),
          Connection.create(429, 279),
          Connection.create(279, 358),
          Connection.create(14, 15),
          Connection.create(15, 316),
          Connection.create(316, 14),
          Connection.create(285, 336),
          Connection.create(336, 9),
          Connection.create(9, 285),
          Connection.create(329, 349),
          Connection.create(349, 350),
          Connection.create(350, 329),
          Connection.create(374, 380),
          Connection.create(380, 252),
          Connection.create(252, 374),
          Connection.create(318, 402),
          Connection.create(402, 403),
          Connection.create(403, 318),
          Connection.create(6, 197),
          Connection.create(197, 419),
          Connection.create(419, 6),
          Connection.create(318, 319),
          Connection.create(319, 325),
          Connection.create(325, 318),
          Connection.create(367, 364),
          Connection.create(364, 365),
          Connection.create(365, 367),
          Connection.create(435, 367),
          Connection.create(367, 397),
          Connection.create(397, 435),
          Connection.create(344, 438),
          Connection.create(438, 439),
          Connection.create(439, 344),
          Connection.create(272, 271),
          Connection.create(271, 311),
          Connection.create(311, 272),
          Connection.create(195, 5),
          Connection.create(5, 281),
          Connection.create(281, 195),
          Connection.create(273, 287),
          Connection.create(287, 291),
          Connection.create(291, 273),
          Connection.create(396, 428),
          Connection.create(428, 199),
          Connection.create(199, 396),
          Connection.create(311, 271),
          Connection.create(271, 268),
          Connection.create(268, 311),
          Connection.create(283, 444),
          Connection.create(444, 445),
          Connection.create(445, 283),
          Connection.create(373, 254),
          Connection.create(254, 339),
          Connection.create(339, 373),
          Connection.create(282, 334),
          Connection.create(334, 296),
          Connection.create(296, 282),
          Connection.create(449, 347),
          Connection.create(347, 346),
          Connection.create(346, 449),
          Connection.create(264, 447),
          Connection.create(447, 454),
          Connection.create(454, 264),
          Connection.create(336, 296),
          Connection.create(296, 299),
          Connection.create(299, 336),
          Connection.create(338, 10),
          Connection.create(10, 151),
          Connection.create(151, 338),
          Connection.create(278, 439),
          Connection.create(439, 455),
          Connection.create(455, 278),
          Connection.create(292, 407),
          Connection.create(407, 415),
          Connection.create(415, 292),
          Connection.create(358, 371),
          Connection.create(371, 355),
          Connection.create(355, 358),
          Connection.create(340, 345),
          Connection.create(345, 372),
          Connection.create(372, 340),
          Connection.create(346, 347),
          Connection.create(347, 280),
          Connection.create(280, 346),
          Connection.create(442, 443),
          Connection.create(443, 282),
          Connection.create(282, 442),
          Connection.create(19, 94),
          Connection.create(94, 370),
          Connection.create(370, 19),
          Connection.create(441, 442),
          Connection.create(442, 295),
          Connection.create(295, 441),
          Connection.create(248, 419),
          Connection.create(419, 197),
          Connection.create(197, 248),
          Connection.create(263, 255),
          Connection.create(255, 359),
          Connection.create(359, 263),
          Connection.create(440, 275),
          Connection.create(275, 274),
          Connection.create(274, 440),
          Connection.create(300, 383),
          Connection.create(383, 368),
          Connection.create(368, 300),
          Connection.create(351, 412),
          Connection.create(412, 465),
          Connection.create(465, 351),
          Connection.create(263, 467),
          Connection.create(467, 466),
          Connection.create(466, 263),
          Connection.create(301, 368),
          Connection.create(368, 389),
          Connection.create(389, 301),
          Connection.create(395, 378),
          Connection.create(378, 379),
          Connection.create(379, 395),
          Connection.create(412, 351),
          Connection.create(351, 419),
          Connection.create(419, 412),
          Connection.create(436, 426),
          Connection.create(426, 322),
          Connection.create(322, 436),
          Connection.create(2, 164),
          Connection.create(164, 393),
          Connection.create(393, 2),
          Connection.create(370, 462),
          Connection.create(462, 461),
          Connection.create(461, 370),
          Connection.create(164, 0),
          Connection.create(0, 267),
          Connection.create(267, 164),
          Connection.create(302, 11),
          Connection.create(11, 12),
          Connection.create(12, 302),
          Connection.create(268, 12),
          Connection.create(12, 13),
          Connection.create(13, 268),
          Connection.create(293, 300),
          Connection.create(300, 301),
          Connection.create(301, 293),
          Connection.create(446, 261),
          Connection.create(261, 340),
          Connection.create(340, 446),
          Connection.create(330, 266),
          Connection.create(266, 425),
          Connection.create(425, 330),
          Connection.create(426, 423),
          Connection.create(423, 391),
          Connection.create(391, 426),
          Connection.create(429, 355),
          Connection.create(355, 437),
          Connection.create(437, 429),
          Connection.create(391, 327),
          Connection.create(327, 326),
          Connection.create(326, 391),
          Connection.create(440, 457),
          Connection.create(457, 438),
          Connection.create(438, 440),
          Connection.create(341, 382),
          Connection.create(382, 362),
          Connection.create(362, 341),
          Connection.create(459, 457),
          Connection.create(457, 461),
          Connection.create(461, 459),
          Connection.create(434, 430),
          Connection.create(430, 394),
          Connection.create(394, 434),
          Connection.create(414, 463),
          Connection.create(463, 362),
          Connection.create(362, 414),
          Connection.create(396, 369),
          Connection.create(369, 262),
          Connection.create(262, 396),
          Connection.create(354, 461),
          Connection.create(461, 457),
          Connection.create(457, 354),
          Connection.create(316, 403),
          Connection.create(403, 402),
          Connection.create(402, 316),
          Connection.create(315, 404),
          Connection.create(404, 403),
          Connection.create(403, 315),
          Connection.create(314, 405),
          Connection.create(405, 404),
          Connection.create(404, 314),
          Connection.create(313, 406),
          Connection.create(406, 405),
          Connection.create(405, 313),
          Connection.create(421, 418),
          Connection.create(418, 406),
          Connection.create(406, 421),
          Connection.create(366, 401),
          Connection.create(401, 361),
          Connection.create(361, 366),
          Connection.create(306, 408),
          Connection.create(408, 407),
          Connection.create(407, 306),
          Connection.create(291, 409),
          Connection.create(409, 408),
          Connection.create(408, 291),
          Connection.create(287, 410),
          Connection.create(410, 409),
          Connection.create(409, 287),
          Connection.create(432, 436),
          Connection.create(436, 410),
          Connection.create(410, 432),
          Connection.create(434, 416),
          Connection.create(416, 411),
          Connection.create(411, 434),
          Connection.create(264, 368),
          Connection.create(368, 383),
          Connection.create(383, 264),
          Connection.create(309, 438),
          Connection.create(438, 457),
          Connection.create(457, 309),
          Connection.create(352, 376),
          Connection.create(376, 401),
          Connection.create(401, 352),
          Connection.create(274, 275),
          Connection.create(275, 4),
          Connection.create(4, 274),
          Connection.create(421, 428),
          Connection.create(428, 262),
          Connection.create(262, 421),
          Connection.create(294, 327),
          Connection.create(327, 358),
          Connection.create(358, 294),
          Connection.create(433, 416),
          Connection.create(416, 367),
          Connection.create(367, 433),
          Connection.create(289, 455),
          Connection.create(455, 439),
          Connection.create(439, 289),
          Connection.create(462, 370),
          Connection.create(370, 326),
          Connection.create(326, 462),
          Connection.create(2, 326),
          Connection.create(326, 370),
          Connection.create(370, 2),
          Connection.create(305, 460),
          Connection.create(460, 455),
          Connection.create(455, 305),
          Connection.create(254, 449),
          Connection.create(449, 448),
          Connection.create(448, 254),
          Connection.create(255, 261),
          Connection.create(261, 446),
          Connection.create(446, 255),
          Connection.create(253, 450),
          Connection.create(450, 449),
          Connection.create(449, 253),
          Connection.create(252, 451),
          Connection.create(451, 450),
          Connection.create(450, 252),
          Connection.create(256, 452),
          Connection.create(452, 451),
          Connection.create(451, 256),
          Connection.create(341, 453),
          Connection.create(453, 452),
          Connection.create(452, 341),
          Connection.create(413, 464),
          Connection.create(464, 463),
          Connection.create(463, 413),
          Connection.create(441, 413),
          Connection.create(413, 414),
          Connection.create(414, 441),
          Connection.create(258, 442),
          Connection.create(442, 441),
          Connection.create(441, 258),
          Connection.create(257, 443),
          Connection.create(443, 442),
          Connection.create(442, 257),
          Connection.create(259, 444),
          Connection.create(444, 443),
          Connection.create(443, 259),
          Connection.create(260, 445),
          Connection.create(445, 444),
          Connection.create(444, 260),
          Connection.create(467, 342),
          Connection.create(342, 445),
          Connection.create(445, 467),
          Connection.create(459, 458),
          Connection.create(458, 250),
          Connection.create(250, 459),
          Connection.create(289, 392),
          Connection.create(392, 290),
          Connection.create(290, 289),
          Connection.create(290, 328),
          Connection.create(328, 460),
          Connection.create(460, 290),
          Connection.create(376, 433),
          Connection.create(433, 435),
          Connection.create(435, 376),
          Connection.create(250, 290),
          Connection.create(290, 392),
          Connection.create(392, 250),
          Connection.create(411, 416),
          Connection.create(416, 433),
          Connection.create(433, 411),
          Connection.create(341, 463),
          Connection.create(463, 464),
          Connection.create(464, 341),
          Connection.create(453, 464),
          Connection.create(464, 465),
          Connection.create(465, 453),
          Connection.create(357, 465),
          Connection.create(465, 412),
          Connection.create(412, 357),
          Connection.create(343, 412),
          Connection.create(412, 399),
          Connection.create(399, 343),
          Connection.create(360, 363),
          Connection.create(363, 440),
          Connection.create(440, 360),
          Connection.create(437, 399),
          Connection.create(399, 456),
          Connection.create(456, 437),
          Connection.create(420, 456),
          Connection.create(456, 363),
          Connection.create(363, 420),
          Connection.create(401, 435),
          Connection.create(435, 288),
          Connection.create(288, 401),
          Connection.create(372, 383),
          Connection.create(383, 353),
          Connection.create(353, 372),
          Connection.create(339, 255),
          Connection.create(255, 249),
          Connection.create(249, 339),
          Connection.create(448, 261),
          Connection.create(261, 255),
          Connection.create(255, 448),
          Connection.create(133, 243),
          Connection.create(243, 190),
          Connection.create(190, 133),
          Connection.create(133, 155),
          Connection.create(155, 112),
          Connection.create(112, 133),
          Connection.create(33, 246),
          Connection.create(246, 247),
          Connection.create(247, 33),
          Connection.create(33, 130),
          Connection.create(130, 25),
          Connection.create(25, 33),
          Connection.create(398, 384),
          Connection.create(384, 286),
          Connection.create(286, 398),
          Connection.create(362, 398),
          Connection.create(398, 414),
          Connection.create(414, 362),
          Connection.create(362, 463),
          Connection.create(463, 341),
          Connection.create(341, 362),
          Connection.create(263, 359),
          Connection.create(359, 467),
          Connection.create(467, 263),
          Connection.create(263, 249),
          Connection.create(249, 255),
          Connection.create(255, 263),
          Connection.create(466, 467),
          Connection.create(467, 260),
          Connection.create(260, 466),
          Connection.create(75, 60),
          Connection.create(60, 166),
          Connection.create(166, 75),
          Connection.create(238, 239),
          Connection.create(239, 79),
          Connection.create(79, 238),
          Connection.create(162, 127),
          Connection.create(127, 139),
          Connection.create(139, 162),
          Connection.create(72, 11),
          Connection.create(11, 37),
          Connection.create(37, 72),
          Connection.create(121, 232),
          Connection.create(232, 120),
          Connection.create(120, 121),
          Connection.create(73, 72),
          Connection.create(72, 39),
          Connection.create(39, 73),
          Connection.create(114, 128),
          Connection.create(128, 47),
          Connection.create(47, 114),
          Connection.create(233, 232),
          Connection.create(232, 128),
          Connection.create(128, 233),
          Connection.create(103, 104),
          Connection.create(104, 67),
          Connection.create(67, 103),
          Connection.create(152, 175),
          Connection.create(175, 148),
          Connection.create(148, 152),
          Connection.create(119, 118),
          Connection.create(118, 101),
          Connection.create(101, 119),
          Connection.create(74, 73),
          Connection.create(73, 40),
          Connection.create(40, 74),
          Connection.create(107, 9),
          Connection.create(9, 108),
          Connection.create(108, 107),
          Connection.create(49, 48),
          Connection.create(48, 131),
          Connection.create(131, 49),
          Connection.create(32, 194),
          Connection.create(194, 211),
          Connection.create(211, 32),
          Connection.create(184, 74),
          Connection.create(74, 185),
          Connection.create(185, 184),
          Connection.create(191, 80),
          Connection.create(80, 183),
          Connection.create(183, 191),
          Connection.create(185, 40),
          Connection.create(40, 186),
          Connection.create(186, 185),
          Connection.create(119, 230),
          Connection.create(230, 118),
          Connection.create(118, 119),
          Connection.create(210, 202),
          Connection.create(202, 214),
          Connection.create(214, 210),
          Connection.create(84, 83),
          Connection.create(83, 17),
          Connection.create(17, 84),
          Connection.create(77, 76),
          Connection.create(76, 146),
          Connection.create(146, 77),
          Connection.create(161, 160),
          Connection.create(160, 30),
          Connection.create(30, 161),
          Connection.create(190, 56),
          Connection.create(56, 173),
          Connection.create(173, 190),
          Connection.create(182, 106),
          Connection.create(106, 194),
          Connection.create(194, 182),
          Connection.create(138, 135),
          Connection.create(135, 192),
          Connection.create(192, 138),
          Connection.create(129, 203),
          Connection.create(203, 98),
          Connection.create(98, 129),
          Connection.create(54, 21),
          Connection.create(21, 68),
          Connection.create(68, 54),
          Connection.create(5, 51),
          Connection.create(51, 4),
          Connection.create(4, 5),
          Connection.create(145, 144),
          Connection.create(144, 23),
          Connection.create(23, 145),
          Connection.create(90, 77),
          Connection.create(77, 91),
          Connection.create(91, 90),
          Connection.create(207, 205),
          Connection.create(205, 187),
          Connection.create(187, 207),
          Connection.create(83, 201),
          Connection.create(201, 18),
          Connection.create(18, 83),
          Connection.create(181, 91),
          Connection.create(91, 182),
          Connection.create(182, 181),
          Connection.create(180, 90),
          Connection.create(90, 181),
          Connection.create(181, 180),
          Connection.create(16, 85),
          Connection.create(85, 17),
          Connection.create(17, 16),
          Connection.create(205, 206),
          Connection.create(206, 36),
          Connection.create(36, 205),
          Connection.create(176, 148),
          Connection.create(148, 140),
          Connection.create(140, 176),
          Connection.create(165, 92),
          Connection.create(92, 39),
          Connection.create(39, 165),
          Connection.create(245, 193),
          Connection.create(193, 244),
          Connection.create(244, 245),
          Connection.create(27, 159),
          Connection.create(159, 28),
          Connection.create(28, 27),
          Connection.create(30, 247),
          Connection.create(247, 161),
          Connection.create(161, 30),
          Connection.create(174, 236),
          Connection.create(236, 196),
          Connection.create(196, 174),
          Connection.create(103, 54),
          Connection.create(54, 104),
          Connection.create(104, 103),
          Connection.create(55, 193),
          Connection.create(193, 8),
          Connection.create(8, 55),
          Connection.create(111, 117),
          Connection.create(117, 31),
          Connection.create(31, 111),
          Connection.create(221, 189),
          Connection.create(189, 55),
          Connection.create(55, 221),
          Connection.create(240, 98),
          Connection.create(98, 99),
          Connection.create(99, 240),
          Connection.create(142, 126),
          Connection.create(126, 100),
          Connection.create(100, 142),
          Connection.create(219, 166),
          Connection.create(166, 218),
          Connection.create(218, 219),
          Connection.create(112, 155),
          Connection.create(155, 26),
          Connection.create(26, 112),
          Connection.create(198, 209),
          Connection.create(209, 131),
          Connection.create(131, 198),
          Connection.create(169, 135),
          Connection.create(135, 150),
          Connection.create(150, 169),
          Connection.create(114, 47),
          Connection.create(47, 217),
          Connection.create(217, 114),
          Connection.create(224, 223),
          Connection.create(223, 53),
          Connection.create(53, 224),
          Connection.create(220, 45),
          Connection.create(45, 134),
          Connection.create(134, 220),
          Connection.create(32, 211),
          Connection.create(211, 140),
          Connection.create(140, 32),
          Connection.create(109, 67),
          Connection.create(67, 108),
          Connection.create(108, 109),
          Connection.create(146, 43),
          Connection.create(43, 91),
          Connection.create(91, 146),
          Connection.create(231, 230),
          Connection.create(230, 120),
          Connection.create(120, 231),
          Connection.create(113, 226),
          Connection.create(226, 247),
          Connection.create(247, 113),
          Connection.create(105, 63),
          Connection.create(63, 52),
          Connection.create(52, 105),
          Connection.create(241, 238),
          Connection.create(238, 242),
          Connection.create(242, 241),
          Connection.create(124, 46),
          Connection.create(46, 156),
          Connection.create(156, 124),
          Connection.create(95, 78),
          Connection.create(78, 96),
          Connection.create(96, 95),
          Connection.create(70, 46),
          Connection.create(46, 63),
          Connection.create(63, 70),
          Connection.create(116, 143),
          Connection.create(143, 227),
          Connection.create(227, 116),
          Connection.create(116, 123),
          Connection.create(123, 111),
          Connection.create(111, 116),
          Connection.create(1, 44),
          Connection.create(44, 19),
          Connection.create(19, 1),
          Connection.create(3, 236),
          Connection.create(236, 51),
          Connection.create(51, 3),
          Connection.create(207, 216),
          Connection.create(216, 205),
          Connection.create(205, 207),
          Connection.create(26, 154),
          Connection.create(154, 22),
          Connection.create(22, 26),
          Connection.create(165, 39),
          Connection.create(39, 167),
          Connection.create(167, 165),
          Connection.create(199, 200),
          Connection.create(200, 208),
          Connection.create(208, 199),
          Connection.create(101, 36),
          Connection.create(36, 100),
          Connection.create(100, 101),
          Connection.create(43, 57),
          Connection.create(57, 202),
          Connection.create(202, 43),
          Connection.create(242, 20),
          Connection.create(20, 99),
          Connection.create(99, 242),
          Connection.create(56, 28),
          Connection.create(28, 157),
          Connection.create(157, 56),
          Connection.create(124, 35),
          Connection.create(35, 113),
          Connection.create(113, 124),
          Connection.create(29, 160),
          Connection.create(160, 27),
          Connection.create(27, 29),
          Connection.create(211, 204),
          Connection.create(204, 210),
          Connection.create(210, 211),
          Connection.create(124, 113),
          Connection.create(113, 46),
          Connection.create(46, 124),
          Connection.create(106, 43),
          Connection.create(43, 204),
          Connection.create(204, 106),
          Connection.create(96, 62),
          Connection.create(62, 77),
          Connection.create(77, 96),
          Connection.create(227, 137),
          Connection.create(137, 116),
          Connection.create(116, 227),
          Connection.create(73, 41),
          Connection.create(41, 72),
          Connection.create(72, 73),
          Connection.create(36, 203),
          Connection.create(203, 142),
          Connection.create(142, 36),
          Connection.create(235, 64),
          Connection.create(64, 240),
          Connection.create(240, 235),
          Connection.create(48, 49),
          Connection.create(49, 64),
          Connection.create(64, 48),
          Connection.create(42, 41),
          Connection.create(41, 74),
          Connection.create(74, 42),
          Connection.create(214, 212),
          Connection.create(212, 207),
          Connection.create(207, 214),
          Connection.create(183, 42),
          Connection.create(42, 184),
          Connection.create(184, 183),
          Connection.create(210, 169),
          Connection.create(169, 211),
          Connection.create(211, 210),
          Connection.create(140, 170),
          Connection.create(170, 176),
          Connection.create(176, 140),
          Connection.create(104, 105),
          Connection.create(105, 69),
          Connection.create(69, 104),
          Connection.create(193, 122),
          Connection.create(122, 168),
          Connection.create(168, 193),
          Connection.create(50, 123),
          Connection.create(123, 187),
          Connection.create(187, 50),
          Connection.create(89, 96),
          Connection.create(96, 90),
          Connection.create(90, 89),
          Connection.create(66, 65),
          Connection.create(65, 107),
          Connection.create(107, 66),
          Connection.create(179, 89),
          Connection.create(89, 180),
          Connection.create(180, 179),
          Connection.create(119, 101),
          Connection.create(101, 120),
          Connection.create(120, 119),
          Connection.create(68, 63),
          Connection.create(63, 104),
          Connection.create(104, 68),
          Connection.create(234, 93),
          Connection.create(93, 227),
          Connection.create(227, 234),
          Connection.create(16, 15),
          Connection.create(15, 85),
          Connection.create(85, 16),
          Connection.create(209, 129),
          Connection.create(129, 49),
          Connection.create(49, 209),
          Connection.create(15, 14),
          Connection.create(14, 86),
          Connection.create(86, 15),
          Connection.create(107, 55),
          Connection.create(55, 9),
          Connection.create(9, 107),
          Connection.create(120, 100),
          Connection.create(100, 121),
          Connection.create(121, 120),
          Connection.create(153, 145),
          Connection.create(145, 22),
          Connection.create(22, 153),
          Connection.create(178, 88),
          Connection.create(88, 179),
          Connection.create(179, 178),
          Connection.create(197, 6),
          Connection.create(6, 196),
          Connection.create(196, 197),
          Connection.create(89, 88),
          Connection.create(88, 96),
          Connection.create(96, 89),
          Connection.create(135, 138),
          Connection.create(138, 136),
          Connection.create(136, 135),
          Connection.create(138, 215),
          Connection.create(215, 172),
          Connection.create(172, 138),
          Connection.create(218, 115),
          Connection.create(115, 219),
          Connection.create(219, 218),
          Connection.create(41, 42),
          Connection.create(42, 81),
          Connection.create(81, 41),
          Connection.create(5, 195),
          Connection.create(195, 51),
          Connection.create(51, 5),
          Connection.create(57, 43),
          Connection.create(43, 61),
          Connection.create(61, 57),
          Connection.create(208, 171),
          Connection.create(171, 199),
          Connection.create(199, 208),
          Connection.create(41, 81),
          Connection.create(81, 38),
          Connection.create(38, 41),
          Connection.create(224, 53),
          Connection.create(53, 225),
          Connection.create(225, 224),
          Connection.create(24, 144),
          Connection.create(144, 110),
          Connection.create(110, 24),
          Connection.create(105, 52),
          Connection.create(52, 66),
          Connection.create(66, 105),
          Connection.create(118, 229),
          Connection.create(229, 117),
          Connection.create(117, 118),
          Connection.create(227, 34),
          Connection.create(34, 234),
          Connection.create(234, 227),
          Connection.create(66, 107),
          Connection.create(107, 69),
          Connection.create(69, 66),
          Connection.create(10, 109),
          Connection.create(109, 151),
          Connection.create(151, 10),
          Connection.create(219, 48),
          Connection.create(48, 235),
          Connection.create(235, 219),
          Connection.create(183, 62),
          Connection.create(62, 191),
          Connection.create(191, 183),
          Connection.create(142, 129),
          Connection.create(129, 126),
          Connection.create(126, 142),
          Connection.create(116, 111),
          Connection.create(111, 143),
          Connection.create(143, 116),
          Connection.create(118, 117),
          Connection.create(117, 50),
          Connection.create(50, 118),
          Connection.create(223, 222),
          Connection.create(222, 52),
          Connection.create(52, 223),
          Connection.create(94, 19),
          Connection.create(19, 141),
          Connection.create(141, 94),
          Connection.create(222, 221),
          Connection.create(221, 65),
          Connection.create(65, 222),
          Connection.create(196, 3),
          Connection.create(3, 197),
          Connection.create(197, 196),
          Connection.create(45, 220),
          Connection.create(220, 44),
          Connection.create(44, 45),
          Connection.create(156, 70),
          Connection.create(70, 139),
          Connection.create(139, 156),
          Connection.create(188, 122),
          Connection.create(122, 245),
          Connection.create(245, 188),
          Connection.create(139, 71),
          Connection.create(71, 162),
          Connection.create(162, 139),
          Connection.create(149, 170),
          Connection.create(170, 150),
          Connection.create(150, 149),
          Connection.create(122, 188),
          Connection.create(188, 196),
          Connection.create(196, 122),
          Connection.create(206, 216),
          Connection.create(216, 92),
          Connection.create(92, 206),
          Connection.create(164, 2),
          Connection.create(2, 167),
          Connection.create(167, 164),
          Connection.create(242, 141),
          Connection.create(141, 241),
          Connection.create(241, 242),
          Connection.create(0, 164),
          Connection.create(164, 37),
          Connection.create(37, 0),
          Connection.create(11, 72),
          Connection.create(72, 12),
          Connection.create(12, 11),
          Connection.create(12, 38),
          Connection.create(38, 13),
          Connection.create(13, 12),
          Connection.create(70, 63),
          Connection.create(63, 71),
          Connection.create(71, 70),
          Connection.create(31, 226),
          Connection.create(226, 111),
          Connection.create(111, 31),
          Connection.create(36, 101),
          Connection.create(101, 205),
          Connection.create(205, 36),
          Connection.create(203, 206),
          Connection.create(206, 165),
          Connection.create(165, 203),
          Connection.create(126, 209),
          Connection.create(209, 217),
          Connection.create(217, 126),
          Connection.create(98, 165),
          Connection.create(165, 97),
          Connection.create(97, 98),
          Connection.create(237, 220),
          Connection.create(220, 218),
          Connection.create(218, 237),
          Connection.create(237, 239),
          Connection.create(239, 241),
          Connection.create(241, 237),
          Connection.create(210, 214),
          Connection.create(214, 169),
          Connection.create(169, 210),
          Connection.create(140, 171),
          Connection.create(171, 32),
          Connection.create(32, 140),
          Connection.create(241, 125),
          Connection.create(125, 237),
          Connection.create(237, 241),
          Connection.create(179, 86),
          Connection.create(86, 178),
          Connection.create(178, 179),
          Connection.create(180, 85),
          Connection.create(85, 179),
          Connection.create(179, 180),
          Connection.create(181, 84),
          Connection.create(84, 180),
          Connection.create(180, 181),
          Connection.create(182, 83),
          Connection.create(83, 181),
          Connection.create(181, 182),
          Connection.create(194, 201),
          Connection.create(201, 182),
          Connection.create(182, 194),
          Connection.create(177, 137),
          Connection.create(137, 132),
          Connection.create(132, 177),
          Connection.create(184, 76),
          Connection.create(76, 183),
          Connection.create(183, 184),
          Connection.create(185, 61),
          Connection.create(61, 184),
          Connection.create(184, 185),
          Connection.create(186, 57),
          Connection.create(57, 185),
          Connection.create(185, 186),
          Connection.create(216, 212),
          Connection.create(212, 186),
          Connection.create(186, 216),
          Connection.create(192, 214),
          Connection.create(214, 187),
          Connection.create(187, 192),
          Connection.create(139, 34),
          Connection.create(34, 156),
          Connection.create(156, 139),
          Connection.create(218, 79),
          Connection.create(79, 237),
          Connection.create(237, 218),
          Connection.create(147, 123),
          Connection.create(123, 177),
          Connection.create(177, 147),
          Connection.create(45, 44),
          Connection.create(44, 4),
          Connection.create(4, 45),
          Connection.create(208, 201),
          Connection.create(201, 32),
          Connection.create(32, 208),
          Connection.create(98, 64),
          Connection.create(64, 129),
          Connection.create(129, 98),
          Connection.create(192, 213),
          Connection.create(213, 138),
          Connection.create(138, 192),
          Connection.create(235, 59),
          Connection.create(59, 219),
          Connection.create(219, 235),
          Connection.create(141, 242),
          Connection.create(242, 97),
          Connection.create(97, 141),
          Connection.create(97, 2),
          Connection.create(2, 141),
          Connection.create(141, 97),
          Connection.create(240, 75),
          Connection.create(75, 235),
          Connection.create(235, 240),
          Connection.create(229, 24),
          Connection.create(24, 228),
          Connection.create(228, 229),
          Connection.create(31, 25),
          Connection.create(25, 226),
          Connection.create(226, 31),
          Connection.create(230, 23),
          Connection.create(23, 229),
          Connection.create(229, 230),
          Connection.create(231, 22),
          Connection.create(22, 230),
          Connection.create(230, 231),
          Connection.create(232, 26),
          Connection.create(26, 231),
          Connection.create(231, 232),
          Connection.create(233, 112),
          Connection.create(112, 232),
          Connection.create(232, 233),
          Connection.create(244, 189),
          Connection.create(189, 243),
          Connection.create(243, 244),
          Connection.create(189, 221),
          Connection.create(221, 190),
          Connection.create(190, 189),
          Connection.create(222, 28),
          Connection.create(28, 221),
          Connection.create(221, 222),
          Connection.create(223, 27),
          Connection.create(27, 222),
          Connection.create(222, 223),
          Connection.create(224, 29),
          Connection.create(29, 223),
          Connection.create(223, 224),
          Connection.create(225, 30),
          Connection.create(30, 224),
          Connection.create(224, 225),
          Connection.create(113, 247),
          Connection.create(247, 225),
          Connection.create(225, 113),
          Connection.create(99, 60),
          Connection.create(60, 240),
          Connection.create(240, 99),
          Connection.create(213, 147),
          Connection.create(147, 215),
          Connection.create(215, 213),
          Connection.create(60, 20),
          Connection.create(20, 166),
          Connection.create(166, 60),
          Connection.create(192, 187),
          Connection.create(187, 213),
          Connection.create(213, 192),
          Connection.create(243, 112),
          Connection.create(112, 244),
          Connection.create(244, 243),
          Connection.create(244, 233),
          Connection.create(233, 245),
          Connection.create(245, 244),
          Connection.create(245, 128),
          Connection.create(128, 188),
          Connection.create(188, 245),
          Connection.create(188, 114),
          Connection.create(114, 174),
          Connection.create(174, 188),
          Connection.create(134, 131),
          Connection.create(131, 220),
          Connection.create(220, 134),
          Connection.create(174, 217),
          Connection.create(217, 236),
          Connection.create(236, 174),
          Connection.create(236, 198),
          Connection.create(198, 134),
          Connection.create(134, 236),
          Connection.create(215, 177),
          Connection.create(177, 58),
          Connection.create(58, 215),
          Connection.create(156, 143),
          Connection.create(143, 124),
          Connection.create(124, 156),
          Connection.create(25, 110),
          Connection.create(110, 7),
          Connection.create(7, 25),
          Connection.create(31, 228),
          Connection.create(228, 25),
          Connection.create(25, 31),
          Connection.create(264, 356),
          Connection.create(356, 368),
          Connection.create(368, 264),
          Connection.create(0, 11),
          Connection.create(11, 267),
          Connection.create(267, 0),
          Connection.create(451, 452),
          Connection.create(452, 349),
          Connection.create(349, 451),
          Connection.create(267, 302),
          Connection.create(302, 269),
          Connection.create(269, 267),
          Connection.create(350, 357),
          Connection.create(357, 277),
          Connection.create(277, 350),
          Connection.create(350, 452),
          Connection.create(452, 357),
          Connection.create(357, 350),
          Connection.create(299, 333),
          Connection.create(333, 297),
          Connection.create(297, 299),
          Connection.create(396, 175),
          Connection.create(175, 377),
          Connection.create(377, 396),
          Connection.create(280, 347),
          Connection.create(347, 330),
          Connection.create(330, 280),
          Connection.create(269, 303),
          Connection.create(303, 270),
          Connection.create(270, 269),
          Connection.create(151, 9),
          Connection.create(9, 337),
          Connection.create(337, 151),
          Connection.create(344, 278),
          Connection.create(278, 360),
          Connection.create(360, 344),
          Connection.create(424, 418),
          Connection.create(418, 431),
          Connection.create(431, 424),
          Connection.create(270, 304),
          Connection.create(304, 409),
          Connection.create(409, 270),
          Connection.create(272, 310),
          Connection.create(310, 407),
          Connection.create(407, 272),
          Connection.create(322, 270),
          Connection.create(270, 410),
          Connection.create(410, 322),
          Connection.create(449, 450),
          Connection.create(450, 347),
          Connection.create(347, 449),
          Connection.create(432, 422),
          Connection.create(422, 434),
          Connection.create(434, 432),
          Connection.create(18, 313),
          Connection.create(313, 17),
          Connection.create(17, 18),
          Connection.create(291, 306),
          Connection.create(306, 375),
          Connection.create(375, 291),
          Connection.create(259, 387),
          Connection.create(387, 260),
          Connection.create(260, 259),
          Connection.create(424, 335),
          Connection.create(335, 418),
          Connection.create(418, 424),
          Connection.create(434, 364),
          Connection.create(364, 416),
          Connection.create(416, 434),
          Connection.create(391, 423),
          Connection.create(423, 327),
          Connection.create(327, 391),
          Connection.create(301, 251),
          Connection.create(251, 298),
          Connection.create(298, 301),
          Connection.create(275, 281),
          Connection.create(281, 4),
          Connection.create(4, 275),
          Connection.create(254, 373),
          Connection.create(373, 253),
          Connection.create(253, 254),
          Connection.create(375, 307),
          Connection.create(307, 321),
          Connection.create(321, 375),
          Connection.create(280, 425),
          Connection.create(425, 411),
          Connection.create(411, 280),
          Connection.create(200, 421),
          Connection.create(421, 18),
          Connection.create(18, 200),
          Connection.create(335, 321),
          Connection.create(321, 406),
          Connection.create(406, 335),
          Connection.create(321, 320),
          Connection.create(320, 405),
          Connection.create(405, 321),
          Connection.create(314, 315),
          Connection.create(315, 17),
          Connection.create(17, 314),
          Connection.create(423, 426),
          Connection.create(426, 266),
          Connection.create(266, 423),
          Connection.create(396, 377),
          Connection.create(377, 369),
          Connection.create(369, 396),
          Connection.create(270, 322),
          Connection.create(322, 269),
          Connection.create(269, 270),
          Connection.create(413, 417),
          Connection.create(417, 464),
          Connection.create(464, 413),
          Connection.create(385, 386),
          Connection.create(386, 258),
          Connection.create(258, 385),
          Connection.create(248, 456),
          Connection.create(456, 419),
          Connection.create(419, 248),
          Connection.create(298, 284),
          Connection.create(284, 333),
          Connection.create(333, 298),
          Connection.create(168, 417),
          Connection.create(417, 8),
          Connection.create(8, 168),
          Connection.create(448, 346),
          Connection.create(346, 261),
          Connection.create(261, 448),
          Connection.create(417, 413),
          Connection.create(413, 285),
          Connection.create(285, 417),
          Connection.create(326, 327),
          Connection.create(327, 328),
          Connection.create(328, 326),
          Connection.create(277, 355),
          Connection.create(355, 329),
          Connection.create(329, 277),
          Connection.create(309, 392),
          Connection.create(392, 438),
          Connection.create(438, 309),
          Connection.create(381, 382),
          Connection.create(382, 256),
          Connection.create(256, 381),
          Connection.create(279, 429),
          Connection.create(429, 360),
          Connection.create(360, 279),
          Connection.create(365, 364),
          Connection.create(364, 379),
          Connection.create(379, 365),
          Connection.create(355, 277),
          Connection.create(277, 437),
          Connection.create(437, 355),
          Connection.create(282, 443),
          Connection.create(443, 283),
          Connection.create(283, 282),
          Connection.create(281, 275),
          Connection.create(275, 363),
          Connection.create(363, 281),
          Connection.create(395, 431),
          Connection.create(431, 369),
          Connection.create(369, 395),
          Connection.create(299, 297),
          Connection.create(297, 337),
          Connection.create(337, 299),
          Connection.create(335, 273),
          Connection.create(273, 321),
          Connection.create(321, 335),
          Connection.create(348, 450),
          Connection.create(450, 349),
          Connection.create(349, 348),
          Connection.create(359, 446),
          Connection.create(446, 467),
          Connection.create(467, 359),
          Connection.create(283, 293),
          Connection.create(293, 282),
          Connection.create(282, 283),
          Connection.create(250, 458),
          Connection.create(458, 462),
          Connection.create(462, 250),
          Connection.create(300, 276),
          Connection.create(276, 383),
          Connection.create(383, 300),
          Connection.create(292, 308),
          Connection.create(308, 325),
          Connection.create(325, 292),
          Connection.create(283, 276),
          Connection.create(276, 293),
          Connection.create(293, 283),
          Connection.create(264, 372),
          Connection.create(372, 447),
          Connection.create(447, 264),
          Connection.create(346, 352),
          Connection.create(352, 340),
          Connection.create(340, 346),
          Connection.create(354, 274),
          Connection.create(274, 19),
          Connection.create(19, 354),
          Connection.create(363, 456),
          Connection.create(456, 281),
          Connection.create(281, 363),
          Connection.create(426, 436),
          Connection.create(436, 425),
          Connection.create(425, 426),
          Connection.create(380, 381),
          Connection.create(381, 252),
          Connection.create(252, 380),
          Connection.create(267, 269),
          Connection.create(269, 393),
          Connection.create(393, 267),
          Connection.create(421, 200),
          Connection.create(200, 428),
          Connection.create(428, 421),
          Connection.create(371, 266),
          Connection.create(266, 329),
          Connection.create(329, 371),
          Connection.create(432, 287),
          Connection.create(287, 422),
          Connection.create(422, 432),
          Connection.create(290, 250),
          Connection.create(250, 328),
          Connection.create(328, 290),
          Connection.create(385, 258),
          Connection.create(258, 384),
          Connection.create(384, 385),
          Connection.create(446, 265),
          Connection.create(265, 342),
          Connection.create(342, 446),
          Connection.create(386, 387),
          Connection.create(387, 257),
          Connection.create(257, 386),
          Connection.create(422, 424),
          Connection.create(424, 430),
          Connection.create(430, 422),
          Connection.create(445, 342),
          Connection.create(342, 276),
          Connection.create(276, 445),
          Connection.create(422, 273),
          Connection.create(273, 424),
          Connection.create(424, 422),
          Connection.create(306, 292),
          Connection.create(292, 307),
          Connection.create(307, 306),
          Connection.create(352, 366),
          Connection.create(366, 345),
          Connection.create(345, 352),
          Connection.create(268, 271),
          Connection.create(271, 302),
          Connection.create(302, 268),
          Connection.create(358, 423),
          Connection.create(423, 371),
          Connection.create(371, 358),
          Connection.create(327, 294),
          Connection.create(294, 460),
          Connection.create(460, 327),
          Connection.create(331, 279),
          Connection.create(279, 294),
          Connection.create(294, 331),
          Connection.create(303, 271),
          Connection.create(271, 304),
          Connection.create(304, 303),
          Connection.create(436, 432),
          Connection.create(432, 427),
          Connection.create(427, 436),
          Connection.create(304, 272),
          Connection.create(272, 408),
          Connection.create(408, 304),
          Connection.create(395, 394),
          Connection.create(394, 431),
          Connection.create(431, 395),
          Connection.create(378, 395),
          Connection.create(395, 400),
          Connection.create(400, 378),
          Connection.create(296, 334),
          Connection.create(334, 299),
          Connection.create(299, 296),
          Connection.create(6, 351),
          Connection.create(351, 168),
          Connection.create(168, 6),
          Connection.create(376, 352),
          Connection.create(352, 411),
          Connection.create(411, 376),
          Connection.create(307, 325),
          Connection.create(325, 320),
          Connection.create(320, 307),
          Connection.create(285, 295),
          Connection.create(295, 336),
          Connection.create(336, 285),
          Connection.create(320, 319),
          Connection.create(319, 404),
          Connection.create(404, 320),
          Connection.create(329, 330),
          Connection.create(330, 349),
          Connection.create(349, 329),
          Connection.create(334, 293),
          Connection.create(293, 333),
          Connection.create(333, 334),
          Connection.create(366, 323),
          Connection.create(323, 447),
          Connection.create(447, 366),
          Connection.create(316, 15),
          Connection.create(15, 315),
          Connection.create(315, 316),
          Connection.create(331, 358),
          Connection.create(358, 279),
          Connection.create(279, 331),
          Connection.create(317, 14),
          Connection.create(14, 316),
          Connection.create(316, 317),
          Connection.create(8, 285),
          Connection.create(285, 9),
          Connection.create(9, 8),
          Connection.create(277, 329),
          Connection.create(329, 350),
          Connection.create(350, 277),
          Connection.create(253, 374),
          Connection.create(374, 252),
          Connection.create(252, 253),
          Connection.create(319, 318),
          Connection.create(318, 403),
          Connection.create(403, 319),
          Connection.create(351, 6),
          Connection.create(6, 419),
          Connection.create(419, 351),
          Connection.create(324, 318),
          Connection.create(318, 325),
          Connection.create(325, 324),
          Connection.create(397, 367),
          Connection.create(367, 365),
          Connection.create(365, 397),
          Connection.create(288, 435),
          Connection.create(435, 397),
          Connection.create(397, 288),
          Connection.create(278, 344),
          Connection.create(344, 439),
          Connection.create(439, 278),
          Connection.create(310, 272),
          Connection.create(272, 311),
          Connection.create(311, 310),
          Connection.create(248, 195),
          Connection.create(195, 281),
          Connection.create(281, 248),
          Connection.create(375, 273),
          Connection.create(273, 291),
          Connection.create(291, 375),
          Connection.create(175, 396),
          Connection.create(396, 199),
          Connection.create(199, 175),
          Connection.create(312, 311),
          Connection.create(311, 268),
          Connection.create(268, 312),
          Connection.create(276, 283),
          Connection.create(283, 445),
          Connection.create(445, 276),
          Connection.create(390, 373),
          Connection.create(373, 339),
          Connection.create(339, 390),
          Connection.create(295, 282),
          Connection.create(282, 296),
          Connection.create(296, 295),
          Connection.create(448, 449),
          Connection.create(449, 346),
          Connection.create(346, 448),
          Connection.create(356, 264),
          Connection.create(264, 454),
          Connection.create(454, 356),
          Connection.create(337, 336),
          Connection.create(336, 299),
          Connection.create(299, 337),
          Connection.create(337, 338),
          Connection.create(338, 151),
          Connection.create(151, 337),
          Connection.create(294, 278),
          Connection.create(278, 455),
          Connection.create(455, 294),
          Connection.create(308, 292),
          Connection.create(292, 415),
          Connection.create(415, 308),
          Connection.create(429, 358),
          Connection.create(358, 355),
          Connection.create(355, 429),
          Connection.create(265, 340),
          Connection.create(340, 372),
          Connection.create(372, 265),
          Connection.create(352, 346),
          Connection.create(346, 280),
          Connection.create(280, 352),
          Connection.create(295, 442),
          Connection.create(442, 282),
          Connection.create(282, 295),
          Connection.create(354, 19),
          Connection.create(19, 370),
          Connection.create(370, 354),
          Connection.create(285, 441),
          Connection.create(441, 295),
          Connection.create(295, 285),
          Connection.create(195, 248),
          Connection.create(248, 197),
          Connection.create(197, 195),
          Connection.create(457, 440),
          Connection.create(440, 274),
          Connection.create(274, 457),
          Connection.create(301, 300),
          Connection.create(300, 368),
          Connection.create(368, 301),
          Connection.create(417, 351),
          Connection.create(351, 465),
          Connection.create(465, 417),
          Connection.create(251, 301),
          Connection.create(301, 389),
          Connection.create(389, 251),
          Connection.create(394, 395),
          Connection.create(395, 379),
          Connection.create(379, 394),
          Connection.create(399, 412),
          Connection.create(412, 419),
          Connection.create(419, 399),
          Connection.create(410, 436),
          Connection.create(436, 322),
          Connection.create(322, 410),
          Connection.create(326, 2),
          Connection.create(2, 393),
          Connection.create(393, 326),
          Connection.create(354, 370),
          Connection.create(370, 461),
          Connection.create(461, 354),
          Connection.create(393, 164),
          Connection.create(164, 267),
          Connection.create(267, 393),
          Connection.create(268, 302),
          Connection.create(302, 12),
          Connection.create(12, 268),
          Connection.create(312, 268),
          Connection.create(268, 13),
          Connection.create(13, 312),
          Connection.create(298, 293),
          Connection.create(293, 301),
          Connection.create(301, 298),
          Connection.create(265, 446),
          Connection.create(446, 340),
          Connection.create(340, 265),
          Connection.create(280, 330),
          Connection.create(330, 425),
          Connection.create(425, 280),
          Connection.create(322, 426),
          Connection.create(426, 391),
          Connection.create(391, 322),
          Connection.create(420, 429),
          Connection.create(429, 437),
          Connection.create(437, 420),
          Connection.create(393, 391),
          Connection.create(391, 326),
          Connection.create(326, 393),
          Connection.create(344, 440),
          Connection.create(440, 438),
          Connection.create(438, 344),
          Connection.create(458, 459),
          Connection.create(459, 461),
          Connection.create(461, 458),
          Connection.create(364, 434),
          Connection.create(434, 394),
          Connection.create(394, 364),
          Connection.create(428, 396),
          Connection.create(396, 262),
          Connection.create(262, 428),
          Connection.create(274, 354),
          Connection.create(354, 457),
          Connection.create(457, 274),
          Connection.create(317, 316),
          Connection.create(316, 402),
          Connection.create(402, 317),
          Connection.create(316, 315),
          Connection.create(315, 403),
          Connection.create(403, 316),
          Connection.create(315, 314),
          Connection.create(314, 404),
          Connection.create(404, 315),
          Connection.create(314, 313),
          Connection.create(313, 405),
          Connection.create(405, 314),
          Connection.create(313, 421),
          Connection.create(421, 406),
          Connection.create(406, 313),
          Connection.create(323, 366),
          Connection.create(366, 361),
          Connection.create(361, 323),
          Connection.create(292, 306),
          Connection.create(306, 407),
          Connection.create(407, 292),
          Connection.create(306, 291),
          Connection.create(291, 408),
          Connection.create(408, 306),
          Connection.create(291, 287),
          Connection.create(287, 409),
          Connection.create(409, 291),
          Connection.create(287, 432),
          Connection.create(432, 410),
          Connection.create(410, 287),
          Connection.create(427, 434),
          Connection.create(434, 411),
          Connection.create(411, 427),
          Connection.create(372, 264),
          Connection.create(264, 383),
          Connection.create(383, 372),
          Connection.create(459, 309),
          Connection.create(309, 457),
          Connection.create(457, 459),
          Connection.create(366, 352),
          Connection.create(352, 401),
          Connection.create(401, 366),
          Connection.create(1, 274),
          Connection.create(274, 4),
          Connection.create(4, 1),
          Connection.create(418, 421),
          Connection.create(421, 262),
          Connection.create(262, 418),
          Connection.create(331, 294),
          Connection.create(294, 358),
          Connection.create(358, 331),
          Connection.create(435, 433),
          Connection.create(433, 367),
          Connection.create(367, 435),
          Connection.create(392, 289),
          Connection.create(289, 439),
          Connection.create(439, 392),
          Connection.create(328, 462),
          Connection.create(462, 326),
          Connection.create(326, 328),
          Connection.create(94, 2),
          Connection.create(2, 370),
          Connection.create(370, 94),
          Connection.create(289, 305),
          Connection.create(305, 455),
          Connection.create(455, 289),
          Connection.create(339, 254),
          Connection.create(254, 448),
          Connection.create(448, 339),
          Connection.create(359, 255),
          Connection.create(255, 446),
          Connection.create(446, 359),
          Connection.create(254, 253),
          Connection.create(253, 449),
          Connection.create(449, 254),
          Connection.create(253, 252),
          Connection.create(252, 450),
          Connection.create(450, 253),
          Connection.create(252, 256),
          Connection.create(256, 451),
          Connection.create(451, 252),
          Connection.create(256, 341),
          Connection.create(341, 452),
          Connection.create(452, 256),
          Connection.create(414, 413),
          Connection.create(413, 463),
          Connection.create(463, 414),
          Connection.create(286, 441),
          Connection.create(441, 414),
          Connection.create(414, 286),
          Connection.create(286, 258),
          Connection.create(258, 441),
          Connection.create(441, 286),
          Connection.create(258, 257),
          Connection.create(257, 442),
          Connection.create(442, 258),
          Connection.create(257, 259),
          Connection.create(259, 443),
          Connection.create(443, 257),
          Connection.create(259, 260),
          Connection.create(260, 444),
          Connection.create(444, 259),
          Connection.create(260, 467),
          Connection.create(467, 445),
          Connection.create(445, 260),
          Connection.create(309, 459),
          Connection.create(459, 250),
          Connection.create(250, 309),
          Connection.create(305, 289),
          Connection.create(289, 290),
          Connection.create(290, 305),
          Connection.create(305, 290),
          Connection.create(290, 460),
          Connection.create(460, 305),
          Connection.create(401, 376),
          Connection.create(376, 435),
          Connection.create(435, 401),
          Connection.create(309, 250),
          Connection.create(250, 392),
          Connection.create(392, 309),
          Connection.create(376, 411),
          Connection.create(411, 433),
          Connection.create(433, 376),
          Connection.create(453, 341),
          Connection.create(341, 464),
          Connection.create(464, 453),
          Connection.create(357, 453),
          Connection.create(453, 465),
          Connection.create(465, 357),
          Connection.create(343, 357),
          Connection.create(357, 412),
          Connection.create(412, 343),
          Connection.create(437, 343),
          Connection.create(343, 399),
          Connection.create(399, 437),
          Connection.create(344, 360),
          Connection.create(360, 440),
          Connection.create(440, 344),
          Connection.create(420, 437),
          Connection.create(437, 456),
          Connection.create(456, 420),
          Connection.create(360, 420),
          Connection.create(420, 363),
          Connection.create(363, 360),
          Connection.create(361, 401),
          Connection.create(401, 288),
          Connection.create(288, 361),
          Connection.create(265, 372),
          Connection.create(372, 353),
          Connection.create(353, 265),
          Connection.create(390, 339),
          Connection.create(339, 249),
          Connection.create(249, 390),
          Connection.create(339, 448),
          Connection.create(448, 255),
          Connection.create(255, 339));

  private FaceMeshConnections() {}
}
